#
# Copyright 2014 OpenStack Foundation
# Copyright 2015 Red Hat, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

import cotyledon
from cotyledon import oslo_config_glue
from oslo_log import log

from aodh import evaluator as evaluator_svc
from aodh import event as event_svc
from aodh import notifier as notifier_svc
from aodh import service

LOG = log.getLogger(__name__)


def notifier():
    sm = cotyledon.ServiceManager()
    conf = service.prepare_service()
    oslo_config_glue.setup(sm, conf)
    sm.add(notifier_svc.AlarmNotifierService,
           workers=conf.notifier.workers, args=(conf,))
    sm.run()


def evaluator():
    sm = cotyledon.ServiceManager()
    conf = service.prepare_service()
    oslo_config_glue.setup(sm, conf)
    sm.add(evaluator_svc.AlarmEvaluationService,
           workers=conf.evaluator.workers, args=(conf,))
    sm.run()


def listener():
    sm = cotyledon.ServiceManager()
    conf = service.prepare_service()
    oslo_config_glue.setup(sm, conf)
    sm.add(event_svc.EventAlarmEvaluationService,
           workers=conf.listener.workers, args=(conf,))
    sm.run()
